/*
VierGewinnt -- Learning Java
Copyright (C) <2006>  Robin and Detlef Wolf
This library is free software -- GNU LGPL license applies.
see http://www.gnu.org/licenses/lgpl.txt for details
*/

public class Robot {
  // compute robot move

  private int maxLevel = 4 ;  // maximum look ahead 
  private VierGewinntField f ;
  private GameLog gameLog ;

  public void setGameLog(GameLog gameLog) {
    this.gameLog = gameLog ;
  }

  public void setMaxLevel(int maxLevel) {
    this.maxLevel = maxLevel ;
    gameLog.println("Computer untersucht ab jetzt bis zu " + maxLevel + " Zge im voraus.") ;
  }

  public int getMaxLevel() {
    return maxLevel ;
  }

  private int getMinWinCol(int columns[]) {  
    // return column with minimal number of moves to win; -1 if no winning move exists
    int minLevel = Integer.MAX_VALUE ;
    int minCol = -1 ;
    for (int c=0; c<f.columnNum; ++c)
      if (columns[c] != -1 && columns[c] < minLevel) {
        minLevel = columns[c] ;
        minCol = c ;
      }
    return minCol ;
  }

  private int getMinWinLevel(int columns[]) {  
    // return lowest # of moves to win, -1 if none exists
    int minLevel = -1 ;
    for (int c=0; c<f.columnNum; ++c)
      if (columns[c] != -1)
        if (minLevel == -1 || columns[c] < minLevel) 
          minLevel = columns[c] ;
    return minLevel ;
  }

  private void printLevels(String message, int columns[]) {  
    gameLog.print(message) ;
    for (int c=0; c<f.columnNum; ++c) 
      if (f.putPossible(c)) 
        gameLog.print(" " + (columns[c]==-1 ? "-" : (""+columns[c]))) ;
      else {
        if (columns[c] != -1) 
          gameLog.print(" [ERROR] ") ;
        gameLog.print(" _") ;
      }
    gameLog.println() ;
  }


  public int computeMove(VierGewinntField f, int who) {
    // robot is playing current who
    // return in which column robot plays, -1 if not possible anymore

    // depending on maxLevel this can be a very time consuming method.
    // This method honors Thread.interrupt(): if the interrupt is raised,
    // this method will return within few milliseconds.

    this.f = f ;

    int column;
    int row ;

    // can computer win for sure?, if yes, do it
    int columns[] = getWinningColumns(who) ;
    printLevels("In diesen Spalten kann der Computer gewinnen:", columns) ;
    int winningColumn = getMinWinCol(columns) ;
    if (winningColumn != -1) {
      try { Thread.currentThread().sleep(200) ; }  // as this goes very fast,
      catch (InterruptedException e) { ;}          // sleep a bit to not frustrate the human
      gameLog.print("Computer zieht in Spalte " + (winningColumn+1)) ;
      if (columns[winningColumn] > 1)
        gameLog.println(" und gewinnt in " + columns[winningColumn] + " Zgen.") ;
      else if (columns[winningColumn] == 1)  
        gameLog.println(" und gewinnt im nchsten Zug.") ;
      else
        gameLog.println(".") ;
      return winningColumn ;
    }
    // no direct path to victory. But can the human win? ...

    // can human win in next move? If yes, answer.
    for (column=0; column<f.columnNum; ++column) 
      if (f.probeCheckWin(column, 3-who) == f.OK_IS_WIN) {
        gameLog.println("Computer mu in Spalte " + (column+1) + " spielen.") ;
        return column;
      }

    // for all moves of the computer, check possible human answers
    for (column=0; column<f.columnNum; ++column) 
      if (f.put(column, who)) {
        columns[column] = getMinWinLevel(getWinningColumns(3-who)) ;
        f.unPut(column) ;
      }
      else
        columns[column] = -1 ;
    // now columns[c]: -1= impossible or OK to move, >0 : human can win in columns[c] moves
    printLevels("Wenn der C. hier spielte, knnte der Mensch gewinnen:", columns) ;

    // are there only loosing moves for the computer?
    int maxCol = -1 ;
    int humanMaxLevel = -1 ;
    boolean hasChance = false ;
    for (int c=0; c<f.columnNum; ++c) 
      if (f.putPossible(c)) {
        if (columns[c] == -1)
          hasChance = true ;
        else {  // human can win in 'c'
          if (columns[c] > humanMaxLevel) {
            humanMaxLevel = columns[c] ;
            maxCol = c ;
          }
        }
      }
    if (!hasChance) {  // if there is no chance, try at least to delay human victory
      gameLog.println("Computer spielt nach " + (maxCol+1) + ". " +
                         "Mensch kann dann in " + columns[maxCol] + " Zgen gewinnen.") ;                       return maxCol ;
    }

    // now: only the move(s) with (columns[c] == -1 && putPossible(c)) are an option.
    // Of those: is there any the human could play and win? Try to prevent this!
    // ... choose the one closest in future
    int humColumns[] = getWinningColumns(3-who) ;
    printLevels("Wenn der Mensch jetzt am Zug wre knnte er hier gewinnen:", humColumns) ;
    int minCol = -1 ;
    int minLevel = Integer.MAX_VALUE ;
    for (int c=0; c<f.columnNum; ++c) 
      if (f.putPossible(c) && columns[c] == -1)      // computer can play here without loosing
        if (humColumns[c] != -1)                   // and human can win in the same place
          if (humColumns[c] < minLevel) {
            minCol = c ;
            minLevel = humColumns[c] ;
          }
    if (minCol != -1) {
      gameLog.println("Computer zieht nach " + (minCol+1) + 
                         " damit Mensch nicht in " + humColumns[minCol] + " Zgen gewinnt.") ;
      return minCol ;
    } 

    // no winning path found, no threat to answer, loosing moves determined, opponent blocked if possible 
    // nows the time for strategy ...
    // currently the strategy is simplistic: play in the middle of uncrowed rows
    // here is a lot of room for improvement, especially in the early phase of the game

    // in other words: ... do anything not directly foolish and avoid columns[c]!=-1 

    // first preferences: a non-loosing stone in the middle of an empty row
    // starting from bottom row, is there an almost empty row? --> if yes, use it
    row = 0 ;
    while (row < f.rowNum) {
      int cnt = 0 ;
      for (column=0; column<f.columnNum; ++column) 
        if (f.field[row][column] == 0)
          ++cnt ;
      if (f.columnNum - cnt < 3)
        break ;
      ++row ;
    }
    if (row < f.rowNum) {  // computer would like to make a move in 'row', in a non-loosing column
      // look for free non-forbidden column, close to middle
      CenterColIter.init(f.columnNum) ;
      while ((column = CenterColIter.next()) != -1) 
        if (f.field[row][column] == 0 && columns[column]==-1) {
          gameLog.println("Computer weiss nicht recht. Spielt in Spalte " + (column+1) + " weil Zeile " + (row+1) + " wenig gefllt ist.") ;
          return column ;
        }
    }   

    // second preferences: a non-loosing stone in the middle
    CenterColIter.init(f.columnNum) ;
    while ((column = CenterColIter.next()) != -1) 
      if (columns[column]==-1 && f.putPossible(column)) {
        gameLog.println("Computer weiss nicht was tun. Spielt in Spalte " + (column+1)) ; 
        return column ;
      }

    // last resort: a stone as close to the middle as possible
    CenterColIter.init(f.columnNum) ;
    while ((column = CenterColIter.next()) != -1) 
      if (f.putPossible(column)) {
        gameLog.println("Verlegenheitszug nach Spalte " + (column+1) + ".") ;
        return column ;
      }

    gameLog.println("Feld ist voll. Kann nicht mehr ziehen. Unentschieden.") ;
    return -1 ;
  }


  private int[] getWinningColumns(int who) {
    // for each allowed computer (who) move determine if all human (3-who) answer moves 
    // allow the computer (who) to win with its next move. Apply this recursively until maxLevel.
    // returns array[0..columNum-1] in which
    //   -1 = not a winning move, not a possible move, or undecidable in the given recursion depth;
    //   values >=0 indicate the max. # of moves needed to win
    return getWinningColumns(who, -1) ;
  }

  private int[] getWinningColumns(int who, int level) {
    // it is assumed that the inital call uses level=-1;
    // in the code below the comment 'computer' refers to the current 'who'
    // and human refers to the current '3-who', i.e. the other player.
    // Recursion decends until 'maxLevel', except when Thread.interrupt() has been
    // raised -- in that case the function behaves like for maxLevel == 0 
    boolean alwaysWin ;
    boolean couldMove ;
    int column ;
    int column2 ;
    int status ;
    int columns[] = new int[f.columnNum] ; // -1 = no win, otherwise level

    ++level ;

    // first check direct wins and initialize columns[]
    for (column=0; column<f.columnNum; ++column)          // computer
      if (f.probeCheckWin(column, who) == f.OK_IS_WIN)   
        columns[column] = level ;            // direct win possible 
      else
        columns[column] = -1 ;

    // if no direct win possible, check deeper ...
    if (level < maxLevel && !Thread.currentThread().isInterrupted()) {
      for (column=0; column<f.columnNum; ++column) {         // computer
        if ((columns[column]) == -1 && f.put(column, who)) { // not a direct computer win && move possible

          int maxWinLevel = -1 ;
          couldMove = false ;
          alwaysWin = true ;  // can human win, now that computer has played in 'column'?
          for (column2=0; column2<f.columnNum; ++column2)    // human
            if (f.probeCheckWin(column2, 3-who) == f.OK_IS_WIN) {
              alwaysWin = false ;            // no point in pursing 'column' any further
              break ;
            }
          // only if there is not direct winning move for the computer, there is need to check further
          if (alwaysWin) {  // can the computer win after every human move?
            for (column2=0; column2<f.columnNum; ++column2) {  // human
              if (f.put(column2, 3-who)) {  // cannot be a winning move for 'human' at this point
                couldMove = true ;
                int wincols[] = getWinningColumns(who, level) ;  // computer
                f.unPut(column2) ;  // undo human move
                int minWinLevel = Integer.MAX_VALUE ;
                for (int c=0; c<f.columnNum; ++c)
                  if (wincols[c] != -1 && wincols[c] < minWinLevel) 
                    minWinLevel = wincols[c] ;
                if (minWinLevel == Integer.MAX_VALUE) {
                  alwaysWin = false ;  // no winning path visible or possible
                  break ;
                }
                if (minWinLevel > maxWinLevel)  // find the biggest minimum
                  maxWinLevel = minWinLevel ;
              }
            }
          }
          f.unPut(column) ;   // undo computer move
          if (couldMove && alwaysWin) 
            columns[column] = maxWinLevel ;
        }
      }        
    }

    return columns ;
  }

}

