/*
VierGewinnt -- Learning Java
Copyright (C) <2006>  Robin and Detlef Wolf
This library is free software -- GNU LGPL license applies.
see http://www.gnu.org/licenses/lgpl.txt for details
*/

// owns the playing field, the robot, the game state (e.g. who's turn), the current message;
// connects to the interface to the human;
// implements the control logic.

import java.util.Vector ;
import java.util.Enumeration ;

public class VierGewinntModel {

  private Robot robot = new Robot() ;
  private HumanMove humanMove ;
  public VierGewinntField f = new VierGewinntField() ;
  GameLog gameLog ;

  private static final int WAIT_FOR_START = 0 ;
  private static final int WAIT_FOR_HUMAN = 1;
  private static final int WAIT_FOR_COMPUTER = 2;
  
  private int state = WAIT_FOR_START ;
  private int who = 1 ; // whos turn is it? 1= Mensch, 2=computer

  private String message = "VierGewinnt" ;

  private Vector observers = new Vector() ;  // who want's to be informed about model changes?

  private String name(int who) {
    return who == 1 ? "Mensch" : "Computer" ;
  }

  public void registerHuman(HumanMove human) {  // called by human UI when initializing
    this.humanMove = human ;
  }

  public void setGameLog(GameLog gameLog) {
    this.gameLog = gameLog ;
    gameLog.println("Wer es zuerst schafft vier Schneeflocken waagerecht, senkrecht") ;
    gameLog.println("oder diagonal anzuordnen gewinnt das Spiel.") ;
    gameLog.println("\nRobin und Detlef Wolf  2006  -- aufbauend auf Java Grundlagen von M.Strahm.") ;
    robot.setGameLog(gameLog) ;
  }

  public void register(ObserverInterface observer) {  // called by interested observers
    observers.add(observer) ;
  }

  private void modelChanged() { // notify all Observers
    for (Enumeration e = observers.elements() ; e.hasMoreElements() ;) 
      ((ObserverInterface)e.nextElement()).modelUpdate() ;
  }

  public void setMessage(String message) {
    this.message = message ;
    modelChanged() ;
  }

  public void setMaxLevel(int maxLevel) {
    robot.setMaxLevel(maxLevel) ;
  }

  public int getMaxLevel() {
    return robot.getMaxLevel() ;
  }


  public String getMessage() {
    return message ;
  }

  public void initialize(int whoBegins) {
    who = whoBegins ;
    f.initialize() ;
  }

  private void beep() {
    java.awt.Toolkit.getDefaultToolkit().beep() ;
  }

  private void beep2() {
    beep() ;
    try { Thread.currentThread().sleep(100) ; }
    catch (InterruptedException e) { ;}
    beep() ;
  }


  InterruptableSwingWorker robotThread = null ;

  public void startMensch() {
    // it may happen that the user presses this button while the robot is computing
    // a move -- tell robot to speed up, but wait until move is completed
    if (robotThread != null)      // send robotThread a flag
      if (robotThread.interruptAndWait()) 
        gameLog.println("Roboter hat Zugberechnung abgebrochen.") ;

    setMessage("Mensch am Zug ...") ;
    initialize(1) ;
    state = WAIT_FOR_HUMAN ;
    humanMove.waitForMove() ;
    modelChanged() ;
  }

  public void startComputer() {
    // it may happen that the user presses this button while the robot is computing
    // a move -- tell robot to speed up, but wait until move is completed
    if (robotThread != null)      // send robotThread a flag
      if (robotThread.interruptAndWait()) 
        gameLog.println("Roboter hat Zugberechnung abgebrochen.") ;

    initialize(2) ;
    state = WAIT_FOR_COMPUTER ;
    move(f.columnNum/2) ;
  }


  public void robotMove() {  // run in robotThread thread to not block GUI updates
    robotThread = new InterruptableSwingWorker() {
      int column ;
      long start = System.currentTimeMillis() ;
      public void timeconsumingFunction() {
        column = robot.computeMove(VierGewinntModel.this.f, 
                                   VierGewinntModel.this.who) ;
      }
      public void finished() {
        System.out.println("Computer brauchte " + 
                           ((float)(System.currentTimeMillis() - start))/1000 + 
                           " Sekunden fr diesen Zug.\n" +                                                       "--------------------------------------------------------------") ;  
        VierGewinntModel.this.move(column) ;
      }
    } ;
    robotThread.start() ;
  }

  public void humanMove(int column) {  // this is invoked from the Human interface only
    gameLog.println("Mensch spielt in Spalte " + (column+1)) ;
    move(column) ;
    if (state == WAIT_FOR_COMPUTER)
      robotMove() ;
  }

  private void move(int column) {
    // current player attempts move in column
    // if move was legal and game is not over, next move() will be done by next player
    // precondtion: state is set
    // postcondition: state has changed if move was possible

    // implementation note: under all circumstances does this function call
    //                      setMessage(), which calls modelChanged()
    //                      to ensure the display is updated

    if (state == WAIT_FOR_START) {
      setMessage("Bitte einen der Startknpfe drcken.") ;
      beep2() ; 
      return ;
    }

    int moveStatus = f.putCheckWin(column, who) ;

    if (moveStatus == f.NOT_POSSIBLE) {
      if (state != WAIT_FOR_HUMAN) 
        System.out.println("program logic error.") ;
      setMessage("Spalte voll. Zug nicht mglich.") ;
      beep2() ; 
      humanMove.waitForMove() ;
      return ;
    }

    if (moveStatus == f.OK_IS_WIN) {
      beep() ;
      state = WAIT_FOR_START ;
      setMessage(name(who) + " hat gewonnen.") ;
      return ;
    }

    // from here on we deal with moveStatus == 'OK_NO_WIN'
    if (f.isTie()) {
      beep() ;
      state = WAIT_FOR_START ;
      setMessage("Unentschieden!") ;
      return ;
    }

    // game continues ... take turns
    who = 3 - who ;
    if (who == 1) {
      state = WAIT_FOR_HUMAN ;
      humanMove.waitForMove() ;
      setMessage("Mensch am Zug ...") ;
    }
    else {
      state = WAIT_FOR_COMPUTER ;
      setMessage("Computer berechnet seinen Zug ... bitte warten.") ;
    }
  }


}

